@@ -20,6 +20,7 @@
 #include "goo/GooHash.h"
 #include "FoFiType1C.h"
 #include "FoFiTrueType.h"
+#include "poppler/Error.h"
 
 //
 // Terminology
@@ -321,10 +322,10 @@ int FoFiTrueType::findCmap(int platform, int encoding) {
   return -1;
 }
 
-Gushort FoFiTrueType::mapCodeToGID(int i, int c) {
+Gushort FoFiTrueType::mapCodeToGID(int i, Guint c) {
   Gushort gid;
-  int segCnt, segEnd, segStart, segDelta, segOffset;
-  int cmapFirst, cmapLen;
+  Guint segCnt, segEnd, segStart, segDelta, segOffset;
+  Guint cmapFirst, cmapLen;
   int pos, a, b, m;
   GBool ok;
 
@@ -335,7 +336,7 @@ Gushort FoFiTrueType::mapCodeToGID(int i, int c) {
   pos = cmaps[i].offset;
   switch (cmaps[i].fmt) {
   case 0:
-    if (c < 0 || c >= cmaps[i].len - 6) {
+    if (c >= cmaps[i].len - 6) {
       return 0;
     }
     gid = getU8(cmaps[i].offset + 6 + c, &ok);
@@ -384,6 +385,31 @@ Gushort FoFiTrueType::mapCodeToGID(int i, int c) {
     }
     gid = getU16BE(pos + 10 + 2 * (c - cmapFirst), &ok);
     break;
+  case 12:
+    segCnt = getU32BE(pos + 12, &ok);
+    a = -1;
+    b = segCnt - 1;
+    segEnd = getU32BE(pos + 16 + 12*b+4, &ok);
+    if (c > segEnd) {
+      return 0;
+    }
+    // invariant: seg[a].end < code <= seg[b].end
+    while (b - a > 1 && ok) {
+      m = (a + b) / 2;
+      segEnd = getU32BE(pos + 16 + 12*m+4, &ok);
+      if (segEnd < c) {
+	a = m;
+      } else {
+	b = m;
+      }
+    }
+    segStart = getU32BE(pos + 16 + 12*b, &ok);
+    segDelta = getU32BE(pos + 16 + 12*b+8, &ok);
+    if (c < segStart) {
+      return 0;
+    }
+    gid = segDelta + (c-segStart);
+    break;
   default:
     return 0;
   }
@@ -912,10 +938,12 @@ void FoFiTrueType::writeTTF(FoFiOutputFunc outputFunc,
   badCmapLen = gFalse;
   cmapLen = 0; // make gcc happy
   if (!missingCmap) {
-    cmapLen = cmaps[0].offset + cmaps[0].len;
-    for (i = 1; i < nCmaps; ++i) {
-      if (cmaps[i].offset + cmaps[i].len > cmapLen) {
-	cmapLen = cmaps[i].offset + cmaps[i].len;
+    if (nCmaps > 0) {
+      cmapLen = cmaps[0].offset + cmaps[0].len;
+      for (i = 1; i < nCmaps; ++i) {
+        if (cmaps[i].offset + cmaps[i].len > cmapLen) {
+	  cmapLen = cmaps[i].offset + cmaps[i].len;
+        }
       }
     }
     cmapLen -= tables[cmapIdx].offset;
@@ -1847,6 +1875,7 @@ void FoFiTrueType::parse() {
   }
   tables = (TrueTypeTable *)gmallocn(nTables, sizeof(TrueTypeTable));
   pos += 12;
+  int wrongTables = 0;
   for (i = 0; i < nTables; ++i) {
     tables[i].tag = getU32BE(pos, &parsedOk);
     tables[i].checksum = getU32BE(pos + 4, &parsedOk);
@@ -1854,10 +1883,14 @@ void FoFiTrueType::parse() {
     tables[i].len = (int)getU32BE(pos + 12, &parsedOk);
     if (tables[i].offset + tables[i].len < tables[i].offset ||
 	tables[i].offset + tables[i].len > len) {
-      parsedOk = gFalse;
+      i--;
+      wrongTables++;
+      error(-1, "Found a bad table definition on true type definition, trying to continue...");
     }
     pos += 16;
   }
+  nTables -= wrongTables;
+  tables = (TrueTypeTable *)greallocn(tables, nTables, sizeof(TrueTypeTable));
   if (!parsedOk) {
     return;
   }
