@@ -27,7 +27,7 @@
 #include "Catalog.h"
 
 //return a newly allocated char* containing an UTF16BE string of size length
-static char* pdfDocEncodingToUTF16 (GooString* orig, int* length)
+char* pdfDocEncodingToUTF16 (GooString* orig, int* length)
 {
   //double size, a unicode char takes 2 char, add 2 for the unicode marker
   *length = 2+2*orig->getLength();
@@ -39,26 +39,28 @@ static char* pdfDocEncodingToUTF16 (GooString* orig, int* length)
   //convert to utf16
   for(int i=2,j=0; i<(*length); i+=2,j++) {
     Unicode u = pdfDocEncoding[(unsigned int)((unsigned char)cstring[j])]&0xffff;
-    result[i+1] = *(char*)(&u);
-    result[i] = *(1+(char*)(&u));
+    result[i] = (u >> 8) & 0xff;
+    result[i+1] = u & 0xff;
   }
   return result;
 }
 
 
 
-FormWidget::FormWidget(XRef *xrefA, Object *aobj, unsigned num, Ref aref) 
+FormWidget::FormWidget(XRef *xrefA, Object *aobj, unsigned num, Ref aref, FormField *fieldA) 
 {
   Object obj1, obj2;
   ref = aref;
   double t;
   ID = 0;
+  defaultsLoaded = gFalse;
   fontSize = 0.0;
   modified = gFalse;
   childNum = num;
   xref = xrefA;
   aobj->copy(&obj);
   type = formUndef;
+  field = fieldA;
   Dict *dict = obj.getDict();
 
   if (!dict->lookup("Rect", &obj1)->isArray()) {
@@ -116,7 +118,7 @@ FormWidget::FormWidget(FormWidget *dest)
   y2 = dest->x2;
 
   type = dest->type;
-
+  field = dest->field;
 }
 
 FormWidget::~FormWidget()
@@ -124,6 +126,11 @@ FormWidget::~FormWidget()
   obj.free ();
 }
 
+bool FormWidget::isReadOnly() const
+{
+  return field->isReadOnly();
+}
+
 int FormWidget::encodeID (unsigned pageNum, unsigned fieldNum)
 {
   return (pageNum << 4*sizeof(unsigned)) + fieldNum;
@@ -135,15 +142,37 @@ void FormWidget::decodeID (unsigned id, unsigned* pageNum, unsigned* fieldNum)
   *fieldNum = (id << 4*sizeof(unsigned)) >> 4*sizeof(unsigned);
 }
 
-FormWidgetButton::FormWidgetButton (XRef *xrefA, Object *aobj, unsigned num, Ref ref, FormFieldButton *p) : FormWidget(xrefA, aobj, num, ref)
+void FormWidget::updateField (const char *key, Object *value)
+{
+  Object *obj1;
+  Ref ref1;
+  Object obj2;
+
+  if (obj.getDict()->lookup ("FT", &obj2)->isName ()) {
+    // It's a composed (annot + field) dict
+    obj1 = &obj;
+    ref1 = ref;
+  } else {
+    // It's an annot dict, we have to modify the Field (parent) dict
+    obj1 = field->getObj ();
+    ref1 = field->getRef ();
+  }
+  obj2.free ();
+
+  obj1->getDict ()->set ("V", value);
+  //notify the xref about the update
+  xref->setModifiedObject(obj1, ref1);
+}
+
+FormWidgetButton::FormWidgetButton (XRef *xrefA, Object *aobj, unsigned num, Ref ref, FormField *p) :
+	FormWidget(xrefA, aobj, num, ref, p)
 {
-  parent = p;
   type = formButton;
+  parent = static_cast<FormFieldButton*>(field);
   onStr = NULL;
-  Dict *dict = obj.getDict();
-  Object obj1;
   state = gFalse;
   siblingsID = NULL;
+  numSiblingsID = 0;
 }
 
 FormWidgetButton::~FormWidgetButton ()
@@ -165,31 +194,33 @@ void FormWidgetButton::setState (GBool astate, GBool calledByParent)
     return;
   //the state modification may be denied by the parent. e.g we don't want to let the user put all combo boxes to false
   if (!calledByParent) { //avoid infinite recursion
+    modified = gTrue;
     if (!parent->setState(childNum, astate)) {
       return;
     }
   }
   state = astate;
+  
   //update appearance
   char *offStr = "Off";
   Object obj1;
   obj1.initName(state?getOnStr():offStr);
-  obj.getDict()->set("V", &obj1);
+  updateField ("V", &obj1);
+
   obj1.initName(state?getOnStr():offStr);
   //modify the Appearance State entry as well
   obj.getDict()->set("AS", &obj1);
-
   //notify the xref about the update
   xref->setModifiedObject(&obj, ref);
 }
 
-bool FormWidgetButton::isReadOnly() const
-{
-  return parent->isReadOnly();
-}
-
 void FormWidgetButton::loadDefaults ()
 {
+  if (defaultsLoaded)
+    return;
+
+  defaultsLoaded = gTrue;
+
   Dict *dict = obj.getDict();
   Object obj1;
 
@@ -215,16 +246,19 @@ void FormWidgetButton::loadDefaults ()
 	      onStr = new GooString (key);
             }
             obj3.free();
+	    if (onStr)
+	      break;
           }
         } else if (obj2.isStream()) {
           Stream *str = obj2.getStream();
           Dict *tmpDict2 = str->getDict();
           Object obj3;
           tmpDict2->lookup("Length", &obj3);
-          int c;
-          onStr = new GooString ("D"); 
+          onStr = new GooString ("D");
         }
         obj2.free();
+	if (onStr)
+	  break;
       }
     }
     obj1.free();
@@ -242,6 +276,7 @@ void FormWidgetButton::loadDefaults ()
   } else if (obj1.isArray()) { //handle the case where we have multiple choices
     error(-1, "FormWidgetButton:: multiple choice isn't supported yet\n");
   }
+  obj1.free();
 }
 
 GBool FormWidgetButton::getState ()
@@ -256,15 +291,20 @@ void FormWidgetButton::setNumSiblingsID (int i)
 }
 
 
-FormWidgetText::FormWidgetText (XRef *xrefA, Object *aobj, unsigned num, Ref ref, FormFieldText *p) : FormWidget(xrefA, aobj, num, ref)
+FormWidgetText::FormWidgetText (XRef *xrefA, Object *aobj, unsigned num, Ref ref, FormField *p) :
+	FormWidget(xrefA, aobj, num, ref, p)
 {
-  parent = p;
   type = formText;
-  Dict *dict = aobj->getDict();
+  parent = static_cast<FormFieldText*>(field);
 }
 
 void FormWidgetText::loadDefaults ()
 {
+  if (defaultsLoaded)
+    return;
+
+  defaultsLoaded = gTrue;
+
   Dict *dict = obj.getDict();
   Object obj1;
 
@@ -335,11 +375,6 @@ bool FormWidgetText::isRichText () const
   return parent->isRichText(); 
 }
 
-bool FormWidgetText::isReadOnly () const
-{
-  return parent->isReadOnly();
-}
-
 int FormWidgetText::getMaxLen () const
 {
   return parent->getMaxLen ();
@@ -352,6 +387,7 @@ void FormWidgetText::setContent(GooString* new_content)
     return;
   }
 
+  modified = gTrue;
   if (new_content == NULL) {
     parent->setContentCopy(NULL);
   } else {
@@ -363,23 +399,27 @@ void FormWidgetText::setContent(GooString* new_content)
     
     GooString *cont = new GooString(new_content);
     parent->setContentCopy(cont);
+    
     Object obj1;
     obj1.initString(cont);
-    obj.getDict()->set("V", &obj1);
-    //notify the xref about the update
-    xref->setModifiedObject(&obj, ref);
-    modified = gTrue;
+    updateField ("V", &obj1);
   }
 }
 
-FormWidgetChoice::FormWidgetChoice(XRef *xrefA, Object *aobj, unsigned num, Ref ref, FormFieldChoice *p) : FormWidget(xrefA, aobj, num, ref)
+FormWidgetChoice::FormWidgetChoice(XRef *xrefA, Object *aobj, unsigned num, Ref ref, FormField *p) :
+	FormWidget(xrefA, aobj, num, ref, p)
 {
-  parent = p;
   type = formChoice;
+  parent = static_cast<FormFieldChoice*>(field);
 }
 
 void FormWidgetChoice::loadDefaults ()
 {
+  if (defaultsLoaded)
+    return;
+
+  defaultsLoaded = gTrue;
+
   Dict *dict = obj.getDict();
   Object obj1;
   if (dict->lookup("Opt", &obj1)->isArray()) {
@@ -443,10 +483,10 @@ void FormWidgetChoice::loadDefaults ()
 
   //convert choice's human readable strings to UTF16
   //and update the /Opt dict entry to reflect this change
-  #ifdef UPDATE_OPT
+#ifdef UPDATE_OPT
   Object *objOpt = new Object();
   objOpt->initArray(xref);
-  #endif
+#endif
   for(int i=0; i<parent->getNumChoices(); i++) {
         if (parent->getChoice(i)->hasUnicodeMarker()) { //string already in UTF16, do nothing
 
@@ -467,10 +507,9 @@ void FormWidgetChoice::loadDefaults ()
     if (tmpCurrentChoice[i])
       parent->select(i);
   }
-  #ifdef UPDATE_OPT
-  obj.getDict()->set("Opt", objOpt);
-  xref->setModifiedObject(&obj, ref); 
-  #endif
+#ifdef UPDATE_OPT
+  updateField ("Opt", objOpt);
+#endif
   delete [] tmpCurrentChoice;
 }
 
@@ -506,9 +545,7 @@ void FormWidgetChoice::_updateV ()
       }
     }
   }
-  obj.getDict()->set("V", &obj1);
-  //notify the xref about the update
-  xref->setModifiedObject(&obj, ref);
+  updateField ("V", &obj1);
   modified = gTrue;
 }
 
@@ -528,6 +565,7 @@ void FormWidgetChoice::select (int i)
     return;
   }
   if (!_checkRange(i)) return;
+  modified = gTrue;
   parent->select(i);
   _updateV();
 }
@@ -539,6 +577,7 @@ void FormWidgetChoice::toggle (int i)
     return;
   }
   if (!_checkRange(i)) return;
+  modified = gTrue;
   parent->toggle(i);
   _updateV();
 }
@@ -549,6 +588,7 @@ void FormWidgetChoice::deselectAll ()
     error(-1, "FormWidgetChoice::deselectAll called on a read only field\n");
     return;
   }
+  modified = gTrue;
   parent->deselectAll();
   _updateV();
 }
@@ -579,6 +619,7 @@ void FormWidgetChoice::setEditChoice (GooString* new_content)
     return;
   }
 
+  modified = gTrue;
   if (new_content == NULL) {
     parent->setEditChoice(NULL);
   } else {
@@ -602,7 +643,6 @@ GooString* FormWidgetChoice::getChoice(int i)
   return parent->getChoice(i); 
 }
 
-
 bool FormWidgetChoice::isCombo () const 
 { 
   return parent->isCombo(); 
@@ -628,34 +668,25 @@ bool FormWidgetChoice::commitOnSelChange () const
   return parent->commitOnSelChange(); 
 }
 
-bool FormWidgetChoice::isReadOnly () const
-{
-  return parent->isReadOnly();
-}
-
 bool FormWidgetChoice::isListBox () const
 {
   return parent->isListBox();
 }
 
-FormWidgetSignature::FormWidgetSignature(XRef *xrefA, Object *aobj, unsigned num, Ref ref, FormFieldSignature *p) : FormWidget(xrefA, aobj, num, ref)
+FormWidgetSignature::FormWidgetSignature(XRef *xrefA, Object *aobj, unsigned num, Ref ref, FormField *p) :
+	FormWidget(xrefA, aobj, num, ref, p)
 {
-  parent = p;
   type = formSignature;
+  parent = static_cast<FormFieldSignature*>(field);
 }
 
-bool FormWidgetSignature::isReadOnly () const
-{
-  return parent->isReadOnly();
-}
 
 //========================================================================
 // FormField
 //========================================================================
 
-FormField::FormField(XRef* xrefA, Object *aobj, const Ref& aref, Form* aform, FormFieldType ty) 
+FormField::FormField(XRef* xrefA, Object *aobj, const Ref& aref, FormFieldType ty) 
 {
-  double t;
   xref = xrefA;
   aobj->copy(&obj);
   Dict* dict = obj.getDict();
@@ -666,7 +697,6 @@ FormField::FormField(XRef* xrefA, Object *aobj, const Ref& aref, Form* aform, Fo
   terminal = false;
   widgets = NULL;
   readOnly = false;
-  form = aform;
   ref = aref;
 
   Object obj1;
@@ -680,6 +710,10 @@ FormField::FormField(XRef* xrefA, Object *aobj, const Ref& aref, Form* aform, Fo
       Object childRef;
       array->get(i, &obj2);
       array->getNF(i, &childRef);
+      if (!obj2.isDict ()) {
+	      error (-1, "Reference to an invalid or non existant object");
+	      continue;
+      }
       //field child
       if (dict->lookup ("FT", &obj3)->isName()) {
         // If I'm not a generic container field and my children
@@ -690,14 +724,14 @@ FormField::FormField(XRef* xrefA, Object *aobj, const Ref& aref, Form* aform, Fo
 	  _createWidget(&obj2, childRef.getRef());
 	}
         obj4.free();
-      } else if(obj2.dictLookup("FT", &obj3)->isName()) {
+      } else if(obj2.dictLookup("FT", &obj3)->isName() || obj2.dictLookup("Kids", &obj3)->isArray()) {
         if(terminal) error(-1, "Field can't have both Widget AND Field as kids\n");
 
         numChildren++;
         children = (FormField**)greallocn(children, numChildren, sizeof(FormField*));
 
         obj3.free();
-        form->createFieldFromDict (&obj2, &children[numChildren-1], xrefA, childRef.getRef());
+	children[numChildren-1] = Form::createFieldFromDict (&obj2, xrefA, childRef.getRef());
       }
       // 1 - we will handle 'collapsed' fields (field + annot in the same dict)
       // as if the annot was in the Kids array of the field
@@ -773,11 +807,20 @@ void FormField::_createWidget (Object *obj, Ref aref)
   numChildren++;
   widgets = (FormWidget**)greallocn(widgets, numChildren, sizeof(FormWidget*));
   //ID = index in "widgets" table
-  if(type==formButton) widgets[numChildren-1] = new FormWidgetButton(xref, obj, numChildren-1, aref, static_cast<FormFieldButton*>(this));
-  else if (type==formText) widgets[numChildren-1] = new FormWidgetText(xref, obj, numChildren-1, aref, static_cast<FormFieldText*>(this));
-  else if (type==formChoice) widgets[numChildren-1] = new FormWidgetChoice(xref, obj, numChildren-1, aref, static_cast<FormFieldChoice*>(this));
-  else if (type==formSignature) widgets[numChildren-1] = new FormWidgetSignature(xref, obj, numChildren-1, aref, static_cast<FormFieldSignature*>(this));
-  else  {
+  switch (type) {
+  case formButton:
+    widgets[numChildren-1] = new FormWidgetButton(xref, obj, numChildren-1, aref, this);
+    break;
+  case formText:
+    widgets[numChildren-1] = new FormWidgetText(xref, obj, numChildren-1, aref, this);
+    break;
+  case formChoice:
+    widgets[numChildren-1] = new FormWidgetChoice(xref, obj, numChildren-1, aref, this);
+    break;
+  case formSignature:
+    widgets[numChildren-1] = new FormWidgetSignature(xref, obj, numChildren-1, aref, this);
+    break;
+  default:
     error(-1, "SubType on non-terminal field, invalid document?");
     numChildren--;
     terminal = false;
@@ -805,10 +848,9 @@ FormWidget* FormField::findWidgetByRef (Ref aref)
 //------------------------------------------------------------------------
 // FormFieldButton
 //------------------------------------------------------------------------
-FormFieldButton::FormFieldButton(XRef *xrefA, Object *aobj, const Ref& ref, Form* form) 
-                                 : FormField(xrefA, aobj, ref, form, formButton)
+FormFieldButton::FormFieldButton(XRef *xrefA, Object *aobj, const Ref& ref) 
+	: FormField(xrefA, aobj, ref, formButton)
 {
-  type = formButton;
   Dict* dict = obj.getDict();
   active_child = -1;
   noAllOff = false;
@@ -878,8 +920,8 @@ GBool FormFieldButton::setState (int num, GBool s)
         if (actChild->getOnStr()) {
           Object obj1;
           obj1.initName(actChild->getOnStr());
-          obj.getDict()->set("V", &obj1);
-          xref->setModifiedObject(&obj, ref);
+	  obj.getDict()->set("V", &obj1);
+	  xref->setModifiedObject(&obj, ref);
         }
       }
     } else {
@@ -900,10 +942,9 @@ FormFieldButton::~FormFieldButton()
 //------------------------------------------------------------------------
 // FormFieldText
 //------------------------------------------------------------------------
-FormFieldText::FormFieldText(XRef *xrefA, Object *aobj, const Ref& ref, Form* form) 
-                             : FormField(xrefA, aobj, ref, form, formText)
+FormFieldText::FormFieldText(XRef *xrefA, Object *aobj, const Ref& ref) 
+	: FormField(xrefA, aobj, ref, formText)
 {
-  type = formText;
   Dict* dict = obj.getDict();
   Object obj1;
   content = NULL;
@@ -958,8 +999,8 @@ FormFieldText::~FormFieldText()
 //------------------------------------------------------------------------
 // FormFieldChoice
 //------------------------------------------------------------------------
-FormFieldChoice::FormFieldChoice(XRef *xrefA, Object *aobj, const Ref& ref, Form* form) 
-                                 : FormField(xrefA, aobj, ref, form, formChoice)
+FormFieldChoice::FormFieldChoice(XRef *xrefA, Object *aobj, const Ref& ref) 
+	: FormField(xrefA, aobj, ref, formChoice)
 {
   numChoices = 0;
   choices = NULL;
@@ -1052,8 +1093,8 @@ void FormFieldChoice::_createChoicesTab ()
 //------------------------------------------------------------------------
 // FormFieldSignature
 //------------------------------------------------------------------------
-FormFieldSignature::FormFieldSignature(XRef *xrefA, Object *dict, const Ref& ref, Form* form)
-                                      : FormField(xrefA, dict, ref, form, formSignature)
+FormFieldSignature::FormFieldSignature(XRef *xrefA, Object *dict, const Ref& ref)
+	: FormField(xrefA, dict, ref, formSignature)
 {
 }
 
@@ -1066,35 +1107,40 @@ FormFieldSignature::~FormFieldSignature()
 // Form
 //------------------------------------------------------------------------
 
-Form::Form(XRef *xrefA, Object* acroForm)
+Form::Form(XRef *xrefA, Object* acroFormA)
 {
-  Array *array = NULL;
   Object obj1;
+
   xref = xrefA;
-  acroForm->dictLookup("Fields",&obj1);
-  if (obj1.isArray()) array = obj1.getArray();
-  obj1.free();
-  if(!array) {
-    error(-1, "Can't get Fields array\n");
-  }
+  acroForm = acroFormA;
+  
   size = 0;
   numFields = 0;
   rootFields = NULL;
-  if (array) {
+
+  acroForm->dictLookup("NeedAppearances", &obj1);
+  needAppearances = (obj1.isBool() && obj1.getBool());
+  obj1.free();
+
+  acroForm->dictLookup("Fields", &obj1);
+  if (obj1.isArray()) {
+    Array *array = obj1.getArray();
+    Object obj2;
+    
     for(int i=0; i<array->getLength(); i++) {
       Object oref;
-      array->get(i, &obj1);
+      array->get(i, &obj2);
       array->getNF(i, &oref);
       if (!oref.isRef()) {
         error(-1, "Direct object in rootFields");
-	obj1.free();
+	obj2.free();
 	oref.free();
         continue;
       }
 
-      if (!obj1.isDict()) {
+      if (!obj2.isDict()) {
         error(-1, "Reference in Fields array to an invalid or non existant object");
-	obj1.free();
+	obj2.free();
 	oref.free();
 	continue;
       }
@@ -1104,21 +1150,24 @@ Form::Form(XRef *xrefA, Object* acroForm)
         rootFields = (FormField**)greallocn(rootFields,size,sizeof(FormField*));
       }
 
-      createFieldFromDict (&obj1, &rootFields[numFields++], xrefA, oref.getRef());
+      rootFields[numFields++] = createFieldFromDict (&obj2, xrefA, oref.getRef());
 
       //Mark readonly field
       Object obj3;
-      if (Form::fieldLookup(obj1.getDict (), "Ff", &obj3)->isInt()) {
+      if (Form::fieldLookup(obj2.getDict (), "Ff", &obj3)->isInt()) {
         int flags = obj3.getInt();
         if (flags & 0x1)
           rootFields[numFields-1]->setReadOnly(true);
       }
       obj3.free();
 
-      obj1.free();
+      obj2.free();
       oref.free();
     }
+  } else {
+    error(-1, "Can't get Fields array\n");
   }
+  obj1.free ();
 }
 
 Form::~Form() {
@@ -1147,23 +1196,27 @@ Object *Form::fieldLookup(Dict *field, char *key, Object *obj) {
   return obj;
 }
 
-void Form::createFieldFromDict (Object* obj, FormField** ptr, XRef *xrefA, const Ref& pref)
+FormField *Form::createFieldFromDict (Object* obj, XRef *xrefA, const Ref& pref)
 {
     Object obj2;
+    FormField *field;
 
     if (Form::fieldLookup(obj->getDict (), "FT", &obj2)->isName("Btn")) {
-      (*ptr) = new FormFieldButton(xrefA, obj, pref, this);
+      field = new FormFieldButton(xrefA, obj, pref);
     } else if (obj2.isName("Tx")) {
-      (*ptr) = new FormFieldText(xrefA, obj, pref, this);
+      field = new FormFieldText(xrefA, obj, pref);
     } else if (obj2.isName("Ch")) {
-      (*ptr) = new FormFieldChoice(xrefA, obj, pref, this);
+      field = new FormFieldChoice(xrefA, obj, pref);
     } else if (obj2.isName("Sig")) {
-      (*ptr) = new FormFieldSignature(xrefA, obj, pref, this);
+      field = new FormFieldSignature(xrefA, obj, pref);
     } else { //we don't have an FT entry => non-terminal field
-      (*ptr) = new FormField(xrefA, obj, pref, this);
+      field = new FormField(xrefA, obj, pref);
     }
     obj2.free();
-    (*ptr)->loadChildrenDefaults();
+    
+    field->loadChildrenDefaults();
+
+    return field;
 }
 
 void Form::postWidgetsLoad ()
@@ -1218,7 +1271,9 @@ FormPageWidgets::FormPageWidgets (XRef *xrefA, Object* annots, unsigned int page
         //create a temporary Annot to get the font size
         Object obj2;
         if (annots->arrayGet(i, &obj2)->isDict()) {
-          Annot* ann = new Annot(xref, NULL ,obj2.getDict(), NULL);
+          Annot *ann;
+	  
+          ann = new Annot(xref, obj2.getDict(), NULL);
           tmp->setFontSize(ann->getFontSize());
           delete ann;
         }
