@@ -955,19 +955,17 @@ void SplashOutputDev::doUpdateFont(GfxState *state) {
   FoFiTrueType *ff;
   Ref embRef;
   Object refObj, strObj;
-  GooString *fileName, *substName;
+  GooString *fileName;
   char *tmpBuf;
   int tmpBufLen;
   Gushort *codeToGID;
   DisplayFontParam *dfp;
-  CharCodeToUnicode *ctu;
   double *textMat;
-  double m11, m12, m21, m22, w1, w2, fontSize;
+  double m11, m12, m21, m22, fontSize;
   SplashCoord mat[4];
-  char *name;
-  Unicode uBuf[8];
-  int c, substIdx, n, code, cmap;
+  int substIdx, n;
   int faceIndex = 0;
+  GBool recreateFont = gFalse;
 
   needFontUpdate = gFalse;
   font = NULL;
@@ -1127,7 +1125,12 @@ void SplashOutputDev::doUpdateFont(GfxState *state) {
 	else
 	  ff = FoFiTrueType::make(tmpBuf, tmpBufLen);
 	if (! ff)
+	{
+	  error(-1, "Couldn't create a font for '%s'",
+	      gfxFont->getName() ? gfxFont->getName()->getCString()
+	                         : "(unnamed)");
 	  goto err2;
+	}
 	codeToGID = ((GfxCIDFont *)gfxFont)->getCodeToGIDMap(ff, &n);
 	delete ff;
       }
@@ -1145,6 +1148,7 @@ void SplashOutputDev::doUpdateFont(GfxState *state) {
       // this shouldn't happen
       goto err2;
     }
+    fontFile->doAdjustMatrix = gTrue;
   }
 
   // get the font matrix
@@ -1160,6 +1164,41 @@ void SplashOutputDev::doUpdateFont(GfxState *state) {
   mat[2] = m21;  mat[3] = m22;
   font = fontEngine->getFont(fontFile, mat, splash->getMatrix());
 
+  // for substituted fonts: adjust the font matrix -- compare the
+  // width of 'm' in the original font and the substituted font
+  if (fontFile->doAdjustMatrix && !gfxFont->isCIDFont()) {
+    double w1, w2;
+    CharCode code;
+    char *name;
+    for (code = 0; code < 256; ++code) {
+      if ((name = ((Gfx8BitFont *)gfxFont)->getCharName(code)) &&
+          name[0] == 'm' && name[1] == '\0') {
+        break;
+      }
+    }
+    if (code < 256) {
+      w1 = ((Gfx8BitFont *)gfxFont)->getWidth(code);
+      w2 = font->getGlyphAdvance(code);
+      if (!gfxFont->isSymbolic() && w2 > 0) {
+        // if real font is substantially narrower than substituted
+        // font, reduce the font size accordingly
+        if (w1 > 0.01 && w1 < 0.9 * w2) {
+          w1 /= w2;
+          m11 *= w1;
+          m21 *= w1;
+          recreateFont = gTrue;
+        }
+      }
+    }
+  }
+
+  if (recreateFont)
+  {
+    mat[0] = m11;  mat[1] = m12;
+    mat[2] = m21;  mat[3] = m22;
+    font = fontEngine->getFont(fontFile, mat, splash->getMatrix());
+  }
+
   if (fontsrc && !fontsrc->isFile)
       fontsrc->unref();
   return;
@@ -2577,9 +2616,9 @@ void SplashOutputDev::setSoftMask(GfxState * /*state*/, double * /*bbox*/,
   if (!alpha && colorMode != splashModeMono1) {
     //~ need to correctly handle the case where no blending color
     //~ space is given
-    tSplash = new Splash(tBitmap, vectorAntialias,
-			 transpGroupStack->origSplash->getScreen());
     if (transpGroupStack->blendingColorSpace) {
+      tSplash = new Splash(tBitmap, vectorAntialias,
+			   transpGroupStack->origSplash->getScreen());
       switch (colorMode) {
       case splashModeMono1:
 	// transparency is not supported in mono1 mode
@@ -2619,8 +2658,12 @@ void SplashOutputDev::setSoftMask(GfxState * /*state*/, double * /*bbox*/,
   memset(softMask->getDataPtr(), 0,
 	 softMask->getRowSize() * softMask->getHeight());
   p = softMask->getDataPtr() + ty * softMask->getRowSize() + tx;
-  for (y = 0; y < tBitmap->getHeight(); ++y) {
-    for (x = 0; x < tBitmap->getWidth(); ++x) {
+  int xMax = tBitmap->getWidth();
+  int yMax = tBitmap->getHeight();
+  if (xMax + tx > bitmap->getWidth()) xMax = bitmap->getWidth() - tx;
+  if (yMax + ty > bitmap->getHeight()) yMax = bitmap->getHeight() - ty;
+  for (y = 0; y < yMax; ++y) {
+    for (x = 0; x < xMax; ++x) {
       tBitmap->getPixel(x, y, color);
       if (alpha) {
 	//~ unimplemented
