@@ -87,6 +87,14 @@ public:
   // This is only used by StreamPredictor.
   virtual int getRawChar();
 
+  // Get next char directly from stream source, without filtering it
+  virtual int getUnfilteredChar () = 0;
+
+  // Resets the stream without reading anything (even not the headers)
+  // WARNING: Reading the stream with something else than getUnfilteredChar 
+  // may lead to unexcepted behaviour until you call reset ()
+  virtual void unfilteredReset () = 0;
+
   // Get next line from stream.
   virtual char *getLine(char *buf, int size);
 
@@ -135,6 +143,65 @@ private:
   int ref;			// reference count
 };
 
+
+ //------------------------------------------------------------------------
+// OutStream
+//
+// This is the base class for all streams that output to a file
+//------------------------------------------------------------------------
+class OutStream {
+public:
+  // Constructor.
+  OutStream ();
+
+  // Desctructor.
+  virtual ~OutStream ();
+
+  // Reference counting.
+  int incRef() { return ++ref; }
+  int decRef() { return --ref; }
+
+  // Close the stream
+  virtual void close() = 0;
+
+  // Return position in stream
+  virtual int getPos() = 0;
+
+  // Put a char in the stream
+  virtual void put (char c) = 0;
+
+  //FIXME
+  // Printf-like function                         2,3 because the first arg is class instance ?
+  virtual void printf (const char *format, ...) = 0 ; //__attribute__((format(printf, 2,3))) = 0;
+
+private:
+  int ref; // reference count
+    
+};
+
+//------------------------------------------------------------------------
+// FileOutStream
+//------------------------------------------------------------------------
+class FileOutStream : public OutStream {
+public:
+  FileOutStream (FILE* fa, Guint startA);
+
+  virtual ~FileOutStream ();
+
+  virtual void close();
+
+  virtual int getPos();
+
+  virtual void put (char c);
+
+  virtual void printf (const char *format, ...);
+private:
+  FILE *f;
+  Guint start;
+
+};
+
+
 //------------------------------------------------------------------------
 // BaseStream
 //
@@ -183,6 +250,9 @@ public:
   virtual Dict *getDict() { return str->getDict(); }
   virtual Stream *getNextStream() { return str; }
 
+  virtual int getUnfilteredChar () { return str->getUnfilteredChar(); }
+  virtual void unfilteredReset () { str->unfilteredReset(); }
+
 protected:
 
   Stream *str;
@@ -289,6 +359,9 @@ public:
   virtual Guint getStart() { return start; }
   virtual void moveStart(int delta);
 
+  virtual int getUnfilteredChar () { return getChar(); }
+  virtual void unfilteredReset () { reset(); }
+
 private:
 
   GBool fillBuf();
@@ -328,6 +401,13 @@ public:
   virtual Guint getStart() { return start; }
   virtual void moveStart(int delta);
 
+  //if needFree = true, the stream will delete buf when it is destroyed
+  //otherwise it will not touch it. Default value is false
+  virtual void setNeedFree (GBool val) { needFree = val; }
+
+  virtual int getUnfilteredChar () { return getChar(); }
+  virtual void unfilteredReset () { reset (); } 
+
 private:
 
   char *buf;
@@ -364,6 +444,10 @@ public:
   virtual Guint getStart();
   virtual void moveStart(int delta);
 
+  virtual int getUnfilteredChar () { return str->getUnfilteredChar(); }
+  virtual void unfilteredReset () { str->unfilteredReset(); }
+
+
 private:
 
   Stream *str;
@@ -512,6 +596,8 @@ public:
   virtual GooString *getPSFilter(int psLevel, char *indent);
   virtual GBool isBinary(GBool last = gTrue);
 
+  virtual void unfilteredReset ();
+
 private:
 
   int encoding;			// 'K' parameter
@@ -587,6 +673,8 @@ public:
   virtual GBool isBinary(GBool last = gTrue);
   Stream *getRawStream() { return str; }
 
+  virtual void unfilteredReset();
+
 private:
 
   GBool progressive;		// set if in progressive mode
@@ -691,6 +779,7 @@ public:
   virtual int getRawChar();
   virtual GooString *getPSFilter(int psLevel, char *indent);
   virtual GBool isBinary(GBool last = gTrue);
+  virtual void unfilteredReset ();
 
 private:
 
