@@ -19,6 +19,7 @@
 #ifndef _POPPLER_LINK_EXTRACTOR_H_
 #define _POPPLER_LINK_EXTRACTOR_H_
 
+#include <Object.h>
 #include <OutputDev.h>
 
 #include <QtCore/QList>
@@ -26,13 +27,13 @@
 namespace Poppler
 {
 
+class Link;
 class PageData;
-class DocumentData;
 
 class LinkExtractorOutputDev : public OutputDev
 {
   public:
-    LinkExtractorOutputDev(PageData *data, DocumentData *doc);
+    LinkExtractorOutputDev(PageData *data);
     virtual ~LinkExtractorOutputDev();
 
     // inherited from OutputDev
@@ -45,9 +46,9 @@ class LinkExtractorOutputDev : public OutputDev
     QList< Link* > links();
 
   private:
-    ::Page *m_popplerPage;
     PageData *m_data;
-    DocumentData *m_doc;
+    double m_pageCropWidth;
+    double m_pageCropHeight;
     QList< Link* > m_links;
 };
 
