@@ -19,16 +19,23 @@
 #ifndef _POPPLER_PAGE_PRIVATE_H_
 #define _POPPLER_PAGE_PRIVATE_H_
 
-#include "poppler-private.h"
+class QRectF;
+
+class LinkAction;
+class Page;
 
 namespace Poppler
 {
 
+class DocumentData;
+class PageTransition;
+
 class PageData {
 public:
-  Link* convertLinkActionToLink(::LinkAction * a, const QRectF &linkArea, DocumentData * doc);
+  Link* convertLinkActionToLink(::LinkAction * a, const QRectF &linkArea);
 
-  const Document *parentDoc;
+  DocumentData *parentDoc;
+  ::Page *page;
   int index;
   PageTransition *transition;
 };
