@@ -17,12 +17,13 @@
  */
 
 #include <poppler-qt4.h>
-#include <QtCore/QFile>
+
 #include <QtCore/QMap>
+#include <QtCore/QVarLengthArray>
 #include <QtGui/QImage>
 #include <QtGui/QPainter>
+
 #include <config.h>
-#include <GlobalParams.h>
 #include <PDFDoc.h>
 #include <Catalog.h>
 #include <Form.h>
@@ -55,7 +56,7 @@ class DummyAnnotation : public Annotation
         virtual SubType subType() const { return A_BASE; }
 };
 
-Link* PageData::convertLinkActionToLink(::LinkAction * a, const QRectF &linkArea, DocumentData * doc)
+Link* PageData::convertLinkActionToLink(::LinkAction * a, const QRectF &linkArea)
 {
   if ( !a )
     return NULL;
@@ -67,7 +68,7 @@ Link* PageData::convertLinkActionToLink(::LinkAction * a, const QRectF &linkArea
     {
       LinkGoTo * g = (LinkGoTo *) a;
       // create link: no ext file, namedDest, object pointer
-      popplerLink = new LinkGoto( linkArea, QString::null, LinkDestination( LinkDestinationData(g->getDest(), g->getNamedDest(), doc ) ) );
+      popplerLink = new LinkGoto( linkArea, QString::null, LinkDestination( LinkDestinationData(g->getDest(), g->getNamedDest(), parentDoc ) ) );
     }
     break;
 
@@ -77,7 +78,7 @@ Link* PageData::convertLinkActionToLink(::LinkAction * a, const QRectF &linkArea
       // copy link file
       const char * fileName = g->getFileName()->getCString();
       // ceate link: fileName, namedDest, object pointer
-      popplerLink = new LinkGoto( linkArea, (QString)fileName, LinkDestination( LinkDestinationData(g->getDest(), g->getNamedDest(), doc ) ) );
+      popplerLink = new LinkGoto( linkArea, (QString)fileName, LinkDestination( LinkDestinationData(g->getDest(), g->getNamedDest(), parentDoc ) ) );
     }
     break;
 
@@ -137,6 +138,7 @@ Link* PageData::convertLinkActionToLink(::LinkAction * a, const QRectF &linkArea
       ::LinkSound *ls = (::LinkSound *)a;
       popplerLink = new LinkSound( linkArea, ls->getVolume(), ls->getSynchronous(), ls->getRepeat(), ls->getMix(), new SoundObject( ls->getSound() ) );
     }
+    break;
 
     case actionMovie:
 /*      TODO this (Movie link)
@@ -157,10 +159,11 @@ Link* PageData::convertLinkActionToLink(::LinkAction * a, const QRectF &linkArea
 }
 
 
-Page::Page(const Document *doc, int index) {
+Page::Page(DocumentData *doc, int index) {
   m_page = new PageData();
   m_page->index = index;
   m_page->parentDoc = doc;
+  m_page->page = doc->doc->getCatalog()->getPage(m_page->index + 1);
   m_page->transition = 0;
 }
 
@@ -174,14 +177,14 @@ QImage Page::renderToImage(double xres, double yres, int x, int y, int w, int h,
 {
   int rotation = (int)rotate * 90;
   QImage img;
-  switch(m_page->parentDoc->m_doc->m_backend)
+  switch(m_page->parentDoc->m_backend)
   {
     case Poppler::Document::SplashBackend:
     {
 #if defined(HAVE_SPLASH)
-      SplashOutputDev *splash_output = static_cast<SplashOutputDev *>(m_page->parentDoc->m_doc->getOutputDev());
+      SplashOutputDev *splash_output = static_cast<SplashOutputDev *>(m_page->parentDoc->getOutputDev());
 
-      m_page->parentDoc->m_doc->doc->displayPageSlice(splash_output, m_page->index + 1, xres, yres,
+      m_page->parentDoc->doc->displayPageSlice(splash_output, m_page->index + 1, xres, yres,
 						 rotation, false, true, false, x, y, w, h);
 
       SplashBitmap *bitmap = splash_output->getBitmap();
@@ -221,15 +224,15 @@ QImage Page::renderToImage(double xres, double yres, int x, int y, int w, int h,
       QImage tmpimg(w == -1 ? qRound( size.width() * xres / 72.0 ) : w, h == -1 ? qRound( size.height() * yres / 72.0 ) : h, QImage::Format_ARGB32);
 
       QPainter painter(&tmpimg);
-      if (m_page->parentDoc->m_doc->m_hints & Document::Antialiasing)
+      if (m_page->parentDoc->m_hints & Document::Antialiasing)
           painter.setRenderHint(QPainter::Antialiasing);
-      if (m_page->parentDoc->m_doc->m_hints & Document::TextAntialiasing)
+      if (m_page->parentDoc->m_hints & Document::TextAntialiasing)
           painter.setRenderHint(QPainter::TextAntialiasing);
       painter.save();
       painter.translate(x == -1 ? 0 : -x, y == -1 ? 0 : -y);
       ArthurOutputDev arthur_output(&painter);
-      arthur_output.startDoc(m_page->parentDoc->m_doc->doc->getXRef());
-      m_page->parentDoc->m_doc->doc->displayPageSlice(&arthur_output,
+      arthur_output.startDoc(m_page->parentDoc->doc->getXRef());
+      m_page->parentDoc->doc->displayPageSlice(&arthur_output,
 						 m_page->index + 1,
 						 xres,
 						 yres,
@@ -258,24 +261,18 @@ QString Page::text(const QRectF &r) const
   GooString *s;
   PDFRectangle *rect;
   QString result;
-  ::Page *p;
   
   output_dev = new TextOutputDev(0, gFalse, gFalse, gFalse);
-  m_page->parentDoc->m_doc->doc->displayPageSlice(output_dev, m_page->index + 1, 72, 72,
+  m_page->parentDoc->doc->displayPageSlice(output_dev, m_page->index + 1, 72, 72,
       0, false, true, false, -1, -1, -1, -1);
-  p = m_page->parentDoc->m_doc->doc->getCatalog()->getPage(m_page->index + 1);
   if (r.isNull())
   {
-    rect = p->getCropBox();
+    rect = m_page->page->getCropBox();
     s = output_dev->getText(rect->x1, rect->y1, rect->x2, rect->y2);
   }
   else
   {
-    double height, y1, y2;
-    height = p->getCropHeight();
-    y1 = height - r.top();
-    y2 = height - r.bottom();
-    s = output_dev->getText(r.left(), y1, r.right(), y2);
+    s = output_dev->getText(r.left(), r.top(), r.right(), r.bottom());
   }
 
   result = QString::fromUtf8(s->getCString());
@@ -307,7 +304,7 @@ bool Page::search(const QString &text, QRectF &rect, SearchDirection direction,
 
   // fetch ourselves a textpage
   TextOutputDev td(NULL, gTrue, gFalse, gFalse);
-  m_page->parentDoc->m_doc->doc->displayPage( &td, m_page->index + 1, 72, 72, rotation, false, true, false );
+  m_page->parentDoc->doc->displayPage( &td, m_page->index + 1, 72, 72, rotation, false, true, false );
   TextPage *textPage=td.takeText();
 
   if (direction == FromTop)
@@ -340,7 +337,7 @@ QList<TextBox*> Page::textList(Rotation rotate) const
   
   int rotation = (int)rotate * 90;
 
-  m_page->parentDoc->m_doc->doc->displayPageSlice(output_dev, m_page->index + 1, 72, 72,
+  m_page->parentDoc->doc->displayPageSlice(output_dev, m_page->index + 1, 72, 72,
       rotation, false, false, false, -1, -1, -1, -1);
 
   TextWordList *word_list = output_dev->makeWordList();
@@ -362,8 +359,12 @@ QList<TextBox*> Page::textList(Rotation rotate) const
     
     TextBox* text_box = new TextBox(string, QRectF(xMin, yMin, xMax-xMin, yMax-yMin));
     text_box->m_data->hasSpaceAfter = word->hasSpaceAfter() == gTrue;
-    text_box->m_data->edge.reserve(word->getLength() + 1);
-    for (int j = 0; j <= word->getLength(); ++j) text_box->m_data->edge.append(word->getEdge(j));
+    text_box->m_data->charBBoxes.reserve(word->getLength());
+    for (int j = 0; j < word->getLength(); ++j)
+    {
+        word->getCharBBox(j, &xMin, &yMin, &xMax, &yMax);
+        text_box->m_data->charBBoxes.append(QRectF(xMin, yMin, xMax-xMin, yMax-yMin));
+    }
     
     wordBoxMap.insert(word, text_box);
     
@@ -387,7 +388,7 @@ PageTransition *Page::transition() const
   if (!m_page->transition) {
     Object o;
     PageTransitionParams params;
-    params.dictObj = m_page->parentDoc->m_doc->doc->getCatalog()->getPage(m_page->index + 1)->getTrans(&o);
+    params.dictObj = m_page->page->getTrans(&o);
     if (params.dictObj->isDict()) m_page->transition = new PageTransition(params);
     o.free();
   }
@@ -398,9 +399,8 @@ Link *Page::action( PageAction act ) const
 {
   if ( act == Page::Opening || act == Page::Closing )
   {
-    ::Page *p = m_page->parentDoc->m_doc->doc->getCatalog()->getPage(m_page->index + 1);
     Object o;
-    p->getActions(&o);
+    m_page->page->getActions(&o);
     if (!o.isDict())
     {
       o.free();
@@ -410,14 +410,14 @@ Link *Page::action( PageAction act ) const
     Object o2;
     const char *key = act == Page::Opening ? "O" : "C";
     dict->lookup((char*)key, &o2);
-    ::LinkAction *act = ::LinkAction::parseAction(&o2, m_page->parentDoc->m_doc->doc->getCatalog()->getBaseURI() );
+    ::LinkAction *lact = ::LinkAction::parseAction(&o2, m_page->parentDoc->doc->getCatalog()->getBaseURI() );
     o2.free();
     o.free();
     Link *popplerLink = NULL;
-    if (act != NULL)
+    if (lact != NULL)
     {
-      popplerLink = m_page->convertLinkActionToLink(act, QRectF(), m_page->parentDoc->m_doc);
-      delete act;
+      popplerLink = m_page->convertLinkActionToLink(lact, QRectF());
+      delete lact;
     }
     return popplerLink;
   }
@@ -426,13 +426,11 @@ Link *Page::action( PageAction act ) const
 
 QSizeF Page::pageSizeF() const
 {
-  ::Page *p;
-  
-  p = m_page->parentDoc->m_doc->doc->getCatalog()->getPage(m_page->index + 1);
-  if ( ( Page::Landscape == orientation() ) || (Page::Seascape == orientation() ) ) {
-      return QSizeF( p->getCropHeight(), p->getCropWidth() );
+  Page::Orientation orient = orientation();
+  if ( ( Page::Landscape == orient ) || (Page::Seascape == orient ) ) {
+      return QSizeF( m_page->page->getCropHeight(), m_page->page->getCropWidth() );
   } else {
-    return QSizeF( p->getCropWidth(), p->getCropHeight() );
+    return QSizeF( m_page->page->getCropWidth(), m_page->page->getCropHeight() );
   }
 }
 
@@ -443,7 +441,7 @@ QSize Page::pageSize() const
 
 Page::Orientation Page::orientation() const
 {
-  int rotation = m_page->parentDoc->m_doc->doc->getCatalog()->getPage(m_page->index + 1)->getRotate();
+  const int rotation = m_page->page->getRotate();
   switch (rotation) {
   case 90:
     return Page::Landscape;
@@ -461,15 +459,13 @@ Page::Orientation Page::orientation() const
 
 void Page::defaultCTM(double *CTM, double dpiX, double dpiY, int rotate, bool upsideDown)
 {
-  ::Page *p;
-  p = m_page->parentDoc->m_doc->doc->getCatalog()->getPage(m_page->index + 1);
-  p->getDefaultCTM(CTM, dpiX, dpiY, rotate, gFalse, upsideDown);
+  m_page->page->getDefaultCTM(CTM, dpiX, dpiY, rotate, gFalse, upsideDown);
 }
 
 QList<Link*> Page::links() const
 {
-  LinkExtractorOutputDev link_dev(m_page, m_page->parentDoc->m_doc);
-  m_page->parentDoc->m_doc->doc->processLinks(&link_dev, m_page->index + 1);
+  LinkExtractorOutputDev link_dev(m_page);
+  m_page->parentDoc->doc->processLinks(&link_dev, m_page->index + 1);
   QList<Link*> popplerLinks = link_dev.links();
 
   return popplerLinks;
@@ -478,7 +474,7 @@ QList<Link*> Page::links() const
 QList<Annotation*> Page::annotations() const
 {
     Object annotArray;
-    ::Page *pdfPage = m_page->parentDoc->m_doc->doc->getCatalog()->getPage(m_page->index + 1);
+    ::Page *pdfPage = m_page->page;
     pdfPage->getAnnots( &annotArray );
     if ( !annotArray.isArray() || annotArray.arrayGetLength() < 1 )
     {
@@ -515,7 +511,7 @@ QList<Annotation*> Page::annotations() const
         annotArray.arrayGet( j, &annot );
         if ( !annot.isDict() )
         {
-            qDebug() << "PDFGenerator: annot not dictionary." << endl;
+            qDebug() << "PDFGenerator: annot not dictionary.";
             annot.free();
             continue;
         }
@@ -531,7 +527,7 @@ QList<Annotation*> Page::annotations() const
         XPDFReader::lookupName( annotDict, "Subtype", subType );
         if ( subType.isEmpty() )
         {
-            qDebug() << "annot has no Subtype" << endl;
+            qDebug() << "Annot has no Subtype";
             annot.free();
             continue;
         }
@@ -618,7 +614,7 @@ QList<Annotation*> Page::annotations() const
             int num = XPDFReader::lookupNumArray( annotDict, (char*)((subType == "Line") ? "L" : "Vertices"), c, 100 );
             if ( num < 4 || (num % 2) != 0 )
             {
-                qDebug() << "L/Vertices wrong fol Line/Poly." << endl;
+                qDebug() << "L/Vertices wrong fol Line/Poly.";
                 delete annotation;
                 annot.free();
                 continue;
@@ -760,7 +756,7 @@ QList<Annotation*> Page::annotations() const
             int num = XPDFReader::lookupNumArray( annotDict, "QuadPoints", c, 80 );
             if ( num < 8 || (num % 8) != 0 )
             {
-                qDebug() << "Wrong QuadPoints for a Highlight annotation." << endl;
+                qDebug() << "Wrong QuadPoints for a Highlight annotation.";
                 delete annotation;
                 annot.free();
                 continue;
@@ -806,7 +802,7 @@ QList<Annotation*> Page::annotations() const
             annotDict->lookup( "InkList", &pathsArray );
             if ( !pathsArray.isArray() || pathsArray.arrayGetLength() < 1 )
             {
-                qDebug() << "InkList not present for ink annot" << endl;
+                qDebug() << "InkList not present for ink annot";
                 delete annotation;
                 annot.free();
                 continue;
@@ -884,7 +880,7 @@ QList<Annotation*> Page::annotations() const
             int num = XPDFReader::lookupNumArray( annotDict, "QuadPoints", c, 8 );
             if ( num > 0 && num != 8 )
             {
-                qDebug() << "Wrong QuadPoints for a Link annotation." << endl;
+                qDebug() << "Wrong QuadPoints for a Link annotation.";
                 delete annotation;
                 annot.free();
                 continue;
@@ -907,8 +903,8 @@ QList<Annotation*> Page::annotations() const
             annotDict->lookup( "PA", &objPA );
             if (!objPA.isNull())
             {
-                ::LinkAction * a = ::LinkAction::parseAction( &objPA, m_page->parentDoc->m_doc->doc->getCatalog()->getBaseURI() );
-                Link * popplerLink = m_page->convertLinkActionToLink( a, QRectF(), m_page->parentDoc->m_doc );
+                ::LinkAction * a = ::LinkAction::parseAction( &objPA, m_page->parentDoc->doc->getCatalog()->getBaseURI() );
+                Link * popplerLink = m_page->convertLinkActionToLink( a, QRectF() );
                 if ( popplerLink )
                 {
                      l->setLinkDestination( popplerLink );
@@ -920,7 +916,7 @@ QList<Annotation*> Page::annotations() const
         {
             // MISSING: Caret, FileAttachment, Sound, Movie, Widget,
             //          Screen, PrinterMark, TrapNet, Watermark, 3D
-            qDebug() << "annotation '" << subType << "' not supported" << endl;
+            qDebug() << "Annotation" << subType << "not supported";
             annot.free();
             continue;
         }
@@ -930,7 +926,7 @@ QList<Annotation*> Page::annotations() const
         double r[4];
         if ( XPDFReader::lookupNumArray( annotDict, "Rect", r, 4 ) != 4 )
         {
-            qDebug() << "Rect is missing for annotation." << endl;
+            qDebug() << "Rect is missing for annotation.";
             annot.free();
             continue;
         }
@@ -1149,7 +1145,7 @@ QList<Annotation*> Page::annotations() const
         if ( addToPage )
         {
             if ( annotationsMap.contains( annotID ) )
-                qDebug() << "PDFGenerator: clash for annotations with ID:" << annotID << endl;
+                qDebug() << "Clash for annotations with ID:" << annotID;
             annotationsMap[ annotID ] = annotation;
         }
     } // end Annotation/PopupWindow parsing loop
@@ -1164,8 +1160,8 @@ QList<Annotation*> Page::annotations() const
             const ResolveWindow & request = *it;
             if ( !popupsMap.contains( request.popupWindowID ) )
                 // warn aboud problems in popup resolving logic
-                qDebug() << "PDFGenerator: can't resolve popup "
-                          << request.popupWindowID << "." << endl;
+                qDebug() << "Cannot resolve popup"
+                          << request.popupWindowID << ".";
             else
             {
                 // set annotation's window properties taking ones from popup
@@ -1199,7 +1195,7 @@ QList<Annotation*> Page::annotations() const
     if ( !resolveRevList.isEmpty() )
     {
         // append children to parents
-        int excludeIDs[ resolveRevList.count() ];   // can't even reach this size
+        QVarLengthArray< int > excludeIDs( resolveRevList.count() );   // can't even reach this size
         int excludeIndex = 0;                       // index in excludeIDs array
         QLinkedList< ResolveRevision >::iterator it = resolveRevList.begin(), end = resolveRevList.end();
         for ( ; it != end; ++it )
@@ -1208,8 +1204,8 @@ QList<Annotation*> Page::annotations() const
             int parentID = request.prevAnnotationID;
             if ( !annotationsMap.contains( parentID ) )
                 // warn about problems in reparenting logic
-                qDebug() << "PDFGenerator: can't reparent annotation to "
-                          << parentID << "." << endl;
+                qDebug() << "Cannot reparent annotation to"
+                          << parentID << ".";
             else
             {
                 // compile and add a Revision structure to the parent annotation
@@ -1262,7 +1258,7 @@ QList<Annotation*> Page::annotations() const
 QList<FormField*> Page::formFields() const
 {
   QList<FormField*> fields;
-  ::Page *p = m_page->parentDoc->m_doc->doc->getCatalog()->getPage(m_page->index + 1);
+  ::Page *p = m_page->page;
   ::FormPageWidgets * form = p->getPageWidgets();
   int formcount = form->getNumWidgets();
   for (int i = 0; i < formcount; ++i)
@@ -1271,15 +1267,21 @@ QList<FormField*> Page::formFields() const
     FormField * ff = NULL;
     switch (fm->getType())
     {
+      case formButton:
+      {
+        ff = new FormFieldButton(m_page->parentDoc, p, static_cast<FormWidgetButton*>(fm));
+      }
+      break;
+
       case formText:
       {
-        ff = new FormFieldText(m_page->parentDoc->m_doc, p, static_cast<FormWidgetText*>(fm));
+        ff = new FormFieldText(m_page->parentDoc, p, static_cast<FormWidgetText*>(fm));
       }
       break;
 
       case formChoice:
       {
-        ff = new FormFieldChoice(m_page->parentDoc->m_doc, p, static_cast<FormWidgetChoice*>(fm));
+        ff = new FormFieldChoice(m_page->parentDoc, p, static_cast<FormWidgetChoice*>(fm));
       }
       break;
 
@@ -1295,16 +1297,13 @@ QList<FormField*> Page::formFields() const
 
 double Page::duration() const
 {
-  ::Page *p;
-  p = m_page->parentDoc->m_doc->doc->getCatalog()->getPage(m_page->index + 1);
-  if (p) return p->getDuration();
-  else return -1;
+  return m_page->page->getDuration();
 }
 
 QString Page::label() const
 {
   GooString goo;
-  if (!m_page->parentDoc->m_doc->doc->getCatalog()->indexToLabel(m_page->index, &goo))
+  if (!m_page->parentDoc->doc->getCatalog()->indexToLabel(m_page->index, &goo))
     return QString();
 
   return QString(goo.getCString());
