@@ -23,6 +23,10 @@
 #ifndef _POPPLER_PRIVATE_H_
 #define _POPPLER_PRIVATE_H_
 
+#include <QtCore/QPointer>
+#include <QtCore/QVariant>
+#include <QtCore/QVector>
+
 #include <config.h>
 #include <GfxState.h>
 #include <GlobalParams.h>
@@ -31,85 +35,27 @@
 #include <PDFDoc.h>
 #include <FontInfo.h>
 #include <OutputDev.h>
+#include <Error.h>
 #if defined(HAVE_SPLASH)
 #include <SplashOutputDev.h>
 #endif
-#include <QtCore/QVariant>
+
+#include "poppler-qt4.h"
 
 class FormWidget;
 
 namespace Poppler {
 
     /* borrowed from kpdf */
-    static QString unicodeToQString(Unicode* u, int len) {
-        QString ret;
-        ret.resize(len);
-        QChar* qch = (QChar*) ret.unicode();
-        for (;len;--len)
-          *qch++ = (QChar) *u++;
-        return ret;
-    }
+    QString unicodeToQString(Unicode* u, int len);
 
-    static GooString *QStringToGooString(const QString &s) {
-        int len = s.length();
-        char *cstring = (char *)gmallocn(s.length(), sizeof(char));
-        for (int i = 0; i < len; ++i)
-          cstring[i] = s.at(i).unicode();
-        GooString *ret = new GooString(cstring, len);
-        gfree(cstring);
-        return ret;
-    }
+    QString UnicodeParsedString(GooString *s1);
 
-    static GooString *QStringToUnicodeGooString(const QString &s) {
-        int len = s.length() * 2 + 2;
-        char *cstring = (char *)gmallocn(len, sizeof(char));
-        cstring[0] = 0xfe;
-        cstring[1] = 0xff;
-        for (int i = 0; i < s.length(); ++i)
-        {
-          cstring[2+i*2] = s.at(i).row();
-          cstring[3+i*2] = s.at(i).cell();
-        }
-        GooString *ret = new GooString(cstring, len);
-        gfree(cstring);
-        return ret;
-    }
+    GooString *QStringToUnicodeGooString(const QString &s);
 
-    static QString UnicodeParsedString(GooString *s1) {
-        if ( !s1 || s1->getLength() == 0 )
-            return QString();
-
-        GBool isUnicode;
-        int i;
-        Unicode u;
-        QString result;
-        if ( ( s1->getChar(0) & 0xff ) == 0xfe && ( s1->getLength() > 1 && ( s1->getChar(1) & 0xff ) == 0xff ) )
-        {
-            isUnicode = gTrue;
-            i = 2;
-        }
-        else
-        {
-            isUnicode = gFalse;
-            i = 0;
-        }
-        while ( i < s1->getLength() )
-        {
-            if ( isUnicode )
-            {
-                u = ( ( s1->getChar(i) & 0xff ) << 8 ) | ( s1->getChar(i+1) & 0xff );
-                i += 2;
-            }
-            else
-            {
-                u = s1->getChar(i) & 0xff;
-                ++i;
-            }
-            result += unicodeToQString( &u, 1 );
-        }
-        return result;
-    }
+    GooString *QStringToGooString(const QString &s);
 
+    void qt4ErrorFunction(int pos, char *msg, va_list args);
 
     class LinkDestinationData
     {
@@ -149,17 +95,23 @@ namespace Poppler {
 		m_outputDev = 0;
 		paperColor = Qt::white;
 		m_hints = 0;
+		m_optContentModel = 0;
 		// It might be more appropriate to delete these in PDFDoc
 		delete ownerPassword;
 		delete userPassword;
 		
-		if ( count == 0 ) globalParams = new GlobalParams();
+		if ( count == 0 )
+		{
+			globalParams = new GlobalParams();
+			setErrorFunction(qt4ErrorFunction);
+		}
 		count ++;
 	    }
 	
 	~DocumentData()
 	{
 		qDeleteAll(m_embeddedFiles);
+		delete (OptContentModel *)m_optContentModel;
 		delete doc;
 		delete m_outputDev;
 		delete m_fontInfoScanner;
@@ -180,9 +132,9 @@ namespace Poppler {
 			{
 #if defined(HAVE_SPLASH)
 			SplashColor bgColor;
-			bgColor[0] = paperColor.red();
+			bgColor[0] = paperColor.blue();
 			bgColor[1] = paperColor.green();
-			bgColor[2] = paperColor.blue();
+			bgColor[2] = paperColor.red();
 			GBool AA = m_hints & Document::TextAntialiasing ? gTrue : gFalse;
 			SplashOutputDev * splashOutputDev = new SplashOutputDev(splashModeXBGR8, 4, gFalse, bgColor, gTrue, AA);
 			splashOutputDev->setVectorAntialias(m_hints & Document::Antialiasing ? gTrue : gFalse);
@@ -272,9 +224,9 @@ namespace Poppler {
 #if defined(HAVE_SPLASH)
 				SplashOutputDev *splash_output = static_cast<SplashOutputDev *>( m_outputDev );
 				SplashColor bgColor;
-				bgColor[0] = paperColor.red();
+				bgColor[0] = paperColor.blue();
 				bgColor[1] = paperColor.green();
-				bgColor[2] = paperColor.blue();
+				bgColor[2] = paperColor.red();
 				splash_output->setPaperColor(bgColor);
 #endif
 				break;
@@ -292,10 +244,11 @@ namespace Poppler {
 			for (int yalv = 0; yalv < numEmb; ++yalv) {
 				EmbFile *ef = doc->getCatalog()->embeddedFile(yalv);
 				m_embeddedFiles.append(new EmbeddedFile(ef));
-				delete ef;
 			}
 		}
 	}
+	
+	static Document *checkDocument(DocumentData *doc);
 
 	PDFDoc *doc;
 	QByteArray fileContents;
@@ -304,6 +257,7 @@ namespace Poppler {
 	Document::RenderBackend m_backend;
 	OutputDev *m_outputDev;
 	QList<EmbeddedFile*> m_embeddedFiles;
+	QPointer<OptContentModel> m_optContentModel;
 	QColor paperColor;
 	int m_hints;
 	static int count;
@@ -332,19 +286,23 @@ namespace Poppler {
 
 		QString fontName;
 		QString fontFile;
-		bool isEmbedded;
-		bool isSubset;
+		bool isEmbedded : 1;
+		bool isSubset : 1;
 		FontInfo::Type type;
     };
 
     class TextBoxData
     {
 	public:
+		TextBoxData()
+		  : nextWord(0), hasSpaceAfter(false)
+		{
+		}
+
 		QString text;
 		QRectF bBox;
 		TextBox *nextWord;
-		QVector<double> edge;	// "near" edge x or y coord of each char
-					//   (plus one extra entry for the last char)
+		QVector<QRectF> charBBoxes; // the boundingRect of each character
 		bool hasSpaceAfter;
     };
 
